@@ -22,11 +22,14 @@ import logging
 import operator
 import os
 from pathlib import Path
+
 from pkg_resources import parse_version
 import ssl
 import types
 import yaml
 
+from irma.shared.utils import aware_datetime
+
 from irma.shared.schemas import (
     apiid,
     ValidationError,
@@ -36,11 +39,14 @@ from irma.shared.schemas.v3 import (
     FileExtSchema,
     FileKioskSchema,
     FileResultSchema,
+    NotificationPageSchema,
     Paginated,
     ProbeSchema,
     ScanRetrievalCodeSchema,
     ScanSchema,
     TagSchema,
+    UserSchema,
+    UserWithPasswordSchema,
 )
 
 
@@ -48,6 +54,9 @@ __all__ = ['AAPI', 'Config', 'IrmaError']
 
 log = logging.getLogger("irma.irmacl")
 logreq = logging.getLogger("irma.irmacl.requests")
+user_schema = UserSchema(exclude=('superuser',))
+user_with_password_schema = UserWithPasswordSchema(
+    partial=('id',), exclude=('superuser',))
 
 
 def auth_guarded(call):
@@ -91,6 +100,10 @@ class IrmaError(Exception):
     pass
 
 
+class IrmaClientError(Exception):
+    pass
+
+
 class Config:
 
     CONF_VENV = Path(os.environ.get("IRMA_CONF", ""))
@@ -173,7 +186,7 @@ class AAPI:
     IRMA_VERSION_MAX = parse_version("v4.0.0a0")
 
     def __init__(
-            self, config=None, *, apicheck=True, limit=None,
+            self, config=None, *, anonymous=False, apicheck=True, limit=None,
             raw=False, loop=None):
         """
         :param config: Config object to use (default None ie.
@@ -196,6 +209,7 @@ class AAPI:
         self.password = config.password
         self.ssl = config.ssl
 
+        self.anonymous = anonymous
         self.apicheck = apicheck
         self.limit = limit
         self.raw = raw
@@ -211,6 +225,9 @@ class AAPI:
         self.session = None
         self.headers = {}
 
+        if self.auth_enabled() and self.anonymous:
+            log.warning("anonymous access prevails over provided credentials")
+
     async def __aenter__(self):
         self.session = await aiohttp.ClientSession(
             timeout=self._timeout, loop=self._loop).__aenter__()
@@ -313,6 +330,47 @@ class AAPI:
 
     _post = auth_guarded(_bare_post)
 
+    async def _bare_put(
+            self, route, session=None, headers=None, data=None, json=None):
+        """ Perform a PUT request on a given route
+
+        :param route: route to query
+        :param session: custom session to use (default None)
+        :param headers: additional headers for the requests (default None)
+        :param data, json: raw or json body to send in the request. Mutually
+            exclusive. (default None)
+        :returns: body of the response
+        :raises: aiohttp.ClientResponseError
+
+        """
+        if session is None:
+            session = self.session
+
+        url = self.url + route
+        if headers is not None:
+            h = self.headers.copy()
+            h.update(headers)
+        else:
+            h = self.headers
+        log.info("PUT %s", url)
+        async with session.put(
+                url, ssl=self.ssl, headers=h, data=data, json=json) as resp:
+            body = (
+                repr(data) if isinstance(data, aiohttp.MultipartWriter) else
+                _json.dumps(data) if data else
+                _json.dumps(json) if json else
+                ''
+            )
+
+            logreq.debug(
+                "%-3s -   PUT - %s - %s - %s - %s",
+                resp.status, self.url, route, body, h
+            )
+            resp.raise_for_status()
+            return await resp.read()
+
+    _put = auth_guarded(_bare_put)
+
     async def _bare_delete(
             self, route, session=None, headers=None, data=None, json=None):
         """ Perform a DELETE request on a given route
@@ -446,6 +504,7 @@ class AAPI:
 
         :param username, password: credentials to use (default None)
         :param session: custom session to use (default None)
+        :raises: IrmaClientError if anonymous access is disabled
         :raises: aiohttp.ClientResponseError
 
         """
@@ -454,19 +513,59 @@ class AAPI:
             "password": password or self.password,
         }
         log.info("Authenticate as %s", data["username"])
-        res = await self._bare_post("/auth/login", session, data=data)
+        if self.anonymous:
+            res = await self._login_anonymous(session)
+        else:
+            try:
+                res = await self._bare_post("/auth/login", session, data=data)
+            except aiohttp.ClientResponseError as e:
+                if e.status == 401:
+                    log.error(
+                        "Cannot authenticate as %s: %s",
+                        data["username"], e.message,
+                    )
+                raise
         res = self._format(res, raw=False)
         self.headers["Authorization"] = "Bearer " + res["token"]
 
+    async def _login_anonymous(self, session=None):
+        """ Authenticate anonymously onto IRMA
+
+        :param session: custom session to use (default None)
+        :returns: body of the response
+        :raises: IrmaClientError if anonymous access is disabled
+        :raises: aiohttp.ConnectionResetError
+
+        """
+        try:
+            return await self._bare_post("/auth/anonymous", session)
+        except aiohttp.ClientResponseError as e:
+            if e.status != 404:
+                raise
+            log.error("Anonymous access disabled server-side, cannot proceed")
+            raise IrmaClientError("Anonymous access disabled")
+
     def auth_enabled(self):
         return bool(self.password)
 
+    @property
+    def users(self):
+        """ Sub-API dedicated to users
+        """
+        return UsersAAPI(self)
+
     @property
     def tags(self):
         """ Sub-API dedicated to tags
         """
         return TagsAAPI(self)
 
+    @property
+    def notifications(self):
+        """ Sub-API dedicated to notifications
+        """
+        return NotificationsAAPI(self)
+
     @property
     def probes(self):
         """ Sub-API dedicated to probes
@@ -542,6 +641,106 @@ class AAPIView:
         return await self.api.__aexit__(exc_type, exc_value, traceback)
 
 
+class UsersAAPI(AAPIView):
+
+    async def list(self, offset=None, limit=None, session=None, raw=None):
+        """ List users
+
+        :param offset: offset for pagination (default None)
+        :param limit: custom limit for pagination (default None)
+        :param session: custom session to use (default None)
+        :param raw: bool, return unprocessed bytes (default None)
+        :returns: the queried page of the list of users
+
+        """
+        query = self._askpage(offset, limit)
+
+        res = await self._get("/users", session, query)
+        schema = Paginated(UserSchema)()
+        return self._format(res, raw, schema=schema)
+
+    async def get(self, user, session=None, raw=None):
+        """ Get user
+        :param user: user or user id
+        :param session: custom session to use (default None)
+        :param raw: bool, return unprocessed bytes (default None)
+        :returns: The queried user
+        """
+        res = await self._get("/users/{}".format(apiid(user)), session)
+        return self._format(res, raw, schema=UserSchema())
+
+    async def new(self, username, password, display_name=None,
+                  permissions=None, session=None, raw=None):
+        """ Create a new user
+
+        :param username, password: credentials to use
+        :param display_name: displayed name in webui (default None)
+        :param permissions: list of permissions (default None)
+        :param session: custom session to use (default None)
+        :param raw: bool, return unprocessed bytes (default None)
+        :returns: a newly created user as UserSchema
+        :raises: aiohttp.ClientResponseError
+            400 Check message for details.
+        """
+        user = {
+            "username": username,
+            "password": password,
+            "display_name": display_name,
+            "permissions": permissions,
+        }
+        user_with_password_schema.validate(user)
+        res = await self._post("/users", session, json=user)
+        return self._format(res, raw, schema=UserSchema())
+
+    async def update(self, user, session=None, raw=None):
+        """ Update user
+        :param user: User object, supports UserWithPasswordSchema
+        :param quiet: bool, do not raise an exception if the tag does not
+            exist (default False)
+        :param session: custom session to use (default None)
+        :returns: modified user as UserSchema
+        """
+        data = user_schema.dump(user)
+        if hasattr(user, 'password'):
+            data['password'] = user.password
+        route = '/users/{}'.format(apiid(user))
+        try:
+            res = await self._put(route, session,
+                                  json=data)
+            return self._format(res, raw, schema=UserSchema())
+        except aiohttp.ClientResponseError as e:
+            if e.status not in [404, 403]:
+                raise
+            log.warning(e)
+            if e.status == 404:
+                raise IrmaError("User {} not found.".format(apiid(user)))
+            if e.status == 403:
+                raise IrmaError("User {} is a system user.".format(
+                    apiid(user)))
+
+    async def delete(self, user, quiet=False, session=None):
+        """ Delete a user
+
+        :param user: user or user_id
+        :param quiet: bool, do not raise an exception if the tag does not
+            exist (default False)
+        :param session: custom session to use (default None)
+        :raises: aiohttp.ClientResponseError
+        :raises: IrmaError if the user does not exist
+
+        """
+        route = '/users/{}'.format(apiid(user))
+        try:
+            await self._delete(route, session)
+        except aiohttp.ClientResponseError as e:
+            if e.status != 404:
+                raise
+            log.warning(e)
+            if not quiet:
+                raise IrmaError("User id: {} does not exist.".format(
+                    apiid(user)))
+
+
 class TagsAAPI(AAPIView):
 
     async def list(self, session=None, raw=None):
@@ -602,6 +801,45 @@ class TagsAAPI(AAPIView):
                 raise IrmaError("Tag {} does not exist.".format(tag))
 
 
+class NotificationsAAPI(AAPIView):
+    async def list(self,  all=False, since=None,
+                   offset=None, limit=None, session=None, raw=None):
+        """ List notifications for the current user.
+            Unless all=True only unread notifications are returned.
+
+        :param all: bool, include already read notifications (default False)
+        :param since: get only notifications emitted after this datetime
+            (default None)
+        :param offset: offset for pagination (default None)
+        :param limit: custom limit for pagination (default None)
+        :param session: custom session to use (default None)
+        :param raw: bool, return unprocessed bytes (default None)
+        :returns: the queried page of the list of notification
+
+        """
+        query = self._askpage(offset, limit)
+        query["all"] = str(all)
+        if since is not None:
+            since = aware_datetime(since)
+            query["since"] = since.isoformat()
+        res = await self._get("/notifications", session, query)
+        return self._format(res, raw, schema=NotificationPageSchema())
+
+    async def update(self, last_read_at, session=None):
+        """ Update the date of the last read notification for the current user
+
+        :param last_read_at: mark notification until this datetime as read
+            Do not use 'now' because you could have new notifications between
+            the time you get notification and the the time you call this
+            function
+        :param session: custom session to use (default None)
+        """
+        data = {
+            "last_read_at": aware_datetime(last_read_at).isoformat()
+        }
+        await self._put("/notifications", session, data=data)
+
+
 class ProbesAAPI(AAPIView):
 
     async def list(self, offset=None, limit=None, session=None, raw=None):
@@ -648,11 +886,15 @@ class FilesAAPI(AAPIView):
         :returns: the queried page of the list of files
 
         """
-        return await self.search(None, None, None, offset, limit, session, raw)
+        return await self.search(
+            offset=offset, limit=limit, session=session, raw=raw,
+        )
 
     async def search(
-            self, type=None, value=None, tags=None, offset=None, limit=None,
-            session=None, raw=None):
+        self, type=None, value=None, since=None, until=None,
+        submitter_types=None, tags=None, sort=None, offset=None, limit=None,
+        session=None, raw=None,
+    ):
         """ List a page of matching results of a query on files
 
         :param type: could be one of
@@ -663,7 +905,15 @@ class FilesAAPI(AAPIView):
         (default "name")
         :param value: the value to search for according to the specified type
         (default None)
+        :param since: datetime after which a file has been scanned (default
+        None)
+        :param until: datetime before which a file has been scanned (default
+        None)
+        :param submitter_types: list of names of the submitters to filter on
+        (default None)
         :param tags: one or a list of tags to filter out results (default None)
+        :param sort: list of sort field with or without direction
+            eg: ['name', 'date:desc'] (default None)
         :param offset: offset for pagination (default None)
         :param limit: custom limit for pagination (default None)
         :param session: custom session to use (default None)
@@ -684,6 +934,20 @@ class FilesAAPI(AAPIView):
                 if isinstance(tags, list)
                 else str(apiid(tags))
             )
+        if sort is not None:
+            query["sort"] = ','.join(sort)
+
+        if since is not None:
+            since = aware_datetime(since)
+            query["since"] = since.isoformat()
+        if until is not None:
+            until = aware_datetime(until)
+            query["until"] = until.isoformat()
+        if since and until and (until <= since):
+            log.warning("Search with a badly ordered time interval")
+
+        if submitter_types is not None:
+            query["submitter_types"] = ",".join(submitter_types)
 
         try:
             res = await self._get("/files", session, query)
@@ -871,21 +1135,23 @@ class FilesAAPI(AAPIView):
     async def rescan(self, limit=None, linger=False, session=None, raw=None):
         """ Start an rescan of all files with outdated probe result
 
-        :param limit: maximum number of task to start (default None)
+        :param limit: maximum number of files to rescan (default None)
         :param linger: bool, wait for the scans to complete (default False)
         :param session: custom session to use (default None)
         :param raw: bool, return unprocessed bytes (default None)
-        :returns: ids of started scans
+        :returns: started scan
         :raises: aiohttp.ClientResponseError
 
         """
         json = None if limit is None else {'limit': limit}
-        res = await self._post("/files/rescan", session, json=json)
+        rescan_json = await self._post("/files/rescan", session, json=json)
+        if rescan_json == b'{}':
+            return
         if linger:
-            for scan_id in self._format(res, raw=False):
-                # wait done sequencially to avoid too many api calls
-                await self.api.scans.waitfor(scan_id, session, raw)
-        return self._format(res, raw)
+            rescan = self._format(rescan_json, raw, schema=ScanSchema())
+            return await self.api.scans.waitfor(rescan, session, raw)
+
+        return self._format(rescan_json, raw, schema=ScanSchema())
 
 
 class ScansAAPI(AAPIView):
